import logging
from typing import Optional

from brownie import chain
from y import convert
from y.constants import weth
from y.datatypes import UsdPrice
from y.decorators import log
from y.networks import Network
from y.prices import magic
from y.typing import AnyAddressType, Block
from y.utils.raw_calls import _totalSupply, raw_call

logger = logging.getLogger(__name__)

@log(logger)
def is_creth(token: AnyAddressType) -> bool:
    address = convert.to_address(token)
    return chain.id == Network.Mainnet and address == '0xcBc1065255cBc3aB41a6868c22d1f1C573AB89fd'

@log(logger)
def get_price_creth(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    address = convert.to_address(token)
    totalBalance = raw_call(address, 'accumulated()', output='int', block=block)
    perShare = totalBalance / _totalSupply(address,block)
    return perShare * magic.get_price(weth,block)



import logging
import os
import threading
from typing import Dict, List, Optional, Tuple

from brownie import chain
from cachetools.func import ttl_cache
from joblib import Parallel, delayed
from multicall import Call, Multicall
from y import convert
from y.datatypes import UsdPrice
from y.decorators import log
from y.exceptions import contract_not_verified
from y.networks import Network
from y.prices.dex.uniswap.v1 import UniswapV1
from y.prices.dex.uniswap.v2 import (NotAUniswapV2Pool, UniswapPoolV2,
                                     UniswapRouterV2)
from y.prices.dex.uniswap.v2_forks import UNISWAPS
from y.typing import Address, AnyAddressType, Block
from y.utils.logging import gh_issue_request
from y.utils.multicall import multicall_same_func_no_input

logger = logging.getLogger(__name__)

# NOTE: If this is failing to pull a price for a token you need, it's likely because that token requires a special swap path.
#       Please add a viable swap path to ..protocols to fetch price data successfully.

class UniswapMultiplexer:
    def __init__(self) -> None:
        self.routers = {}
        for name in UNISWAPS:
            try: self.routers[name] = UniswapRouterV2(UNISWAPS[name]['router'])
            except ValueError as e: # TODO do this better
                if not contract_not_verified(e):
                    raise
        self.factories = [UNISWAPS[name]['factory'] for name in UNISWAPS]
        self.v1 = UniswapV1()
        self._uid_lock = threading.Lock()

    @log(logger)
    def is_uniswap_pool(self, token_address: AnyAddressType) -> bool:
        token_address = convert.to_address(token_address)
        try:
            pool = UniswapPoolV2(token_address)
            is_pool = all(pool.get_pool_details())
            if is_pool and pool.factory not in self.factories:
                gh_issue_request(f'UniClone Factory {pool.factory} is unknown to ypricemagic.', logger)
                self.factories.append(pool.factory)
            return is_pool

        except NotAUniswapV2Pool: return False

    @log(logger)
    def get_price_v1(self, token_address: Address, block: Optional[Block] = None) -> UsdPrice:
        return self.v1.get_price(token_address, block)
    
    @log(logger)
    @ttl_cache(ttl=600)
    def lp_price(self, token_address: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
        """ Get Uniswap/Sushiswap LP token price. """
        return UniswapPoolV2(token_address).get_price(block=block)
    
    @log(logger)
    @ttl_cache(ttl=36000)
    def get_price(self, token_in: AnyAddressType, block: Optional[Block] = None, protocol: Optional[str] = None) -> Optional[UsdPrice]:
        """
        Calculate a price based on Uniswap Router quote for selling one `token_in`.
        Always finds the deepest swap path for `token_in`.
        """
        token_in = convert.to_address(token_in)

        if protocol:
            return self.routers[protocol].get_price(token_in, block=block)

        for router in self.routers_by_depth(token_in, block=block):
            # tries each known router from most to least liquid
            # returns the first price we get back, almost always from the deepest router
            price = router.get_price(token_in, block=block)
            if price:
                return price
        
        if chain.id == Network.Mainnet:
            return self.get_price_v1(token_in, block)
        
        return None
    

    @log(logger)
    def deepest_router(self, token_in: AnyAddressType, block: Optional[Block] = None) -> Optional[UniswapRouterV2]:
        token_in = convert.to_address(token_in)

        for router in self.routers_by_depth(token_in, block=block):
            return router # will return first router in the dict, or None if no supported routers
        return None


    @log(logger)
    def routers_by_depth(self, token_in: AnyAddressType, block: Optional[Block] = None) -> Dict[UniswapRouterV2,str]:
        '''
        Returns a dict {router: pool} ordered by liquidity depth, greatest to least
        '''
        calls = self._routers_by_depth_calls(token_in)
        reserves = Multicall(calls, block_id=block, require_success=False)().values()
        return self._routers_by_depth_logic(token_in, reserves)


    def _routers_by_depth_calls(self, token_in: AnyAddressType) -> List[Call]:
        ''' Returns List[Calls] to use for multicall and bool to indicate whether to require success on the multicall. '''
        token_in = convert.to_address(token_in)

        pools_to_routers = {pool: router for router in self.routers.values() for pool in router.pools_for_token(token_in)}
        return [Call(pool, 'getReserves()((uint112,uint112,uint32))', [[self._next_uid(), None]]) for pool in pools_to_routers]
    

    def _routers_by_depth_logic(self, token_in: AnyAddressType, getReserves_call_responses: List[Tuple[int,int,int]]) -> Dict[UniswapRouterV2, str]:
        ''' uses the results of a multicall built by `_routers_by_debth_calls` to output routers_by_depth. '''
        pools_to_routers = {pool: router for router in self.routers.values() for pool in router.pools_for_token(token_in)}

        routers_by_depth = {}
        for router, pool, reserves in zip(pools_to_routers.values(), pools_to_routers.keys(), getReserves_call_responses):
            if reserves is None:
                continue
            if token_in == router.pools[pool]['token0']:
                routers_by_depth[reserves[0]] = {router: pool}
            elif token_in == router.pools[pool]['token1']:
                routers_by_depth[reserves[1]] = {router: pool}
        return {router: pool for balance in sorted(routers_by_depth, reverse=True) for router, pool in routers_by_depth[balance].items()}

    def _next_uid(self) -> str:
        with self._uid_lock:
            try:
                self._last_uid += 1
            except AttributeError:
                self._last_uid = 0
            return f"call_{self._last_uid}"

    def _get_prices(
        self,
        token_addresses: List[AnyAddressType],
        block: Optional[Block] = None,
        dop: int = int(os.environ.get('DOP',4))
        ) -> List[Optional[UsdPrice]]:
        ''' Get uniswap prices for multiple tokens at once, optimized for speed. '''

        token_addresses = [convert.to_address(token) for token in token_addresses]
        if block is None:
            block = chain.height

        # First we collect calls so we can evaluate `routers_by_depth` efficiently using multicall, for minimal total calls.
        calls_dict = {
            i: self._routers_by_depth_calls(token_addresses[i])
            for i in range(len(token_addresses))
        }

        # Then we set up and execute the multicall.
        calls = [call for i, calls in calls_dict.items() for call in calls]
        responses = list(Multicall(calls, block_id=block, require_success=False)().values())
        assert len(calls) == len(responses), f"There is an issue with your multicall. Must investigate. len(calls): {len(calls)}  len(reserves): {len(responses)}"

        # Then we separate the results by input token and evaluate `routers_by_depth` logic with the results.
        routers_by_depth = {}
        done = 0
        for i, calls in calls_dict.items():
            start = done
            end = start + len(calls)
            reserves = responses[start:end]
            routers_by_depth[i] = self._routers_by_depth_logic(token_addresses[i],reserves)
            done = end

        # Get prices for any token that has one
        # TODO optimize this for speed
        prices = Parallel(dop, 'threading')(delayed(self._get_price_from_routers)(token_addresses[i], routers, block) for i, routers in routers_by_depth.items())
        assert len(token_addresses) == len(prices)
        
        return prices

    def _get_price_from_routers(self, token: AnyAddressType, routers: Dict[UniswapRouterV2,str], block: Optional[Block] = None) -> Optional[UsdPrice]:
        token = convert.to_address(token)
        for router in routers:
            price = router.get_price(token, block)
            if price:
                return price

uniswap_multiplexer = UniswapMultiplexer()



import logging
import os
from collections import Counter, defaultdict
from itertools import zip_longest
from typing import Any, Dict, List, Optional

from brownie import chain, web3
from brownie.convert.datatypes import EthAddress
from brownie.network.event import EventDict, _decode_logs
from eth_typing import ChecksumAddress
from joblib import Parallel, delayed
from toolz import groupby
from web3.middleware.filter import block_ranges
from web3.types import LogReceipt
from y.contracts import contract_creation_block
from y.decorators import auto_retry
from y.typing import Address, Block
from y.utils.cache import memory
from y.utils.middleware import BATCH_SIZE

logger = logging.getLogger(__name__)


def decode_logs(logs: List[LogReceipt]) -> EventDict:
    """
    Decode logs to events and enrich them with additional info.
    """
    decoded = _decode_logs(logs)
    for i, log in enumerate(logs):
        setattr(decoded[i], "block_number", log["blockNumber"])
        setattr(decoded[i], "transaction_hash", log["transactionHash"])
        setattr(decoded[i], "log_index", log["logIndex"])
    return decoded


@auto_retry
def create_filter(address, topics=None):
    """
    Create a log filter for one or more contracts.
    Set fromBlock as the earliest creation block.
    """
    if isinstance(address, list):
        start_block = min(map(contract_creation_block, address))
    else:
        start_block = contract_creation_block(address)

    return web3.eth.filter({"address": address, "fromBlock": start_block, "topics": topics})


@auto_retry
def get_logs_asap(address: Optional[Address], topics: Optional[List[str]], from_block: Optional[Block] = None, to_block: Optional[Block] = None, verbose: int = 0) -> List[Any]:
    logs = []

    if from_block is None:
        from_block = 0 if address is None else contract_creation_block(address)
    if to_block is None:
        to_block = chain.height

    ranges = list(block_ranges(from_block, to_block, BATCH_SIZE))
    if verbose > 0:
        logger.info('fetching %d batches', len(ranges))
    
    batches = Parallel(int(os.environ.get('DOP',8)), "threading", verbose=verbose)(delayed(_get_logs)(address, topics, start, end) for start, end in ranges)
    
    for batch in batches:
        logs.extend(batch)
        del batch
    del batches

    return logs


def logs_to_balance_checkpoints(logs) -> Dict[EthAddress,int]:
    """
    Convert Transfer logs to `{address: {from_block: balance}}` checkpoints.
    """
    balances = Counter()
    checkpoints = defaultdict(dict)
    for block, block_logs in groupby('blockNumber', logs).items():
        events = decode_logs(block_logs)
        for log in events:
            # ZERO_ADDRESS tracks -totalSupply
            sender, receiver, amount = log.values()  # there can be several different aliases
            balances[sender] -= amount
            checkpoints[sender][block] = balances[sender]
            balances[receiver] += amount
            checkpoints[receiver][block] = balances[receiver]
    return checkpoints


def checkpoints_to_weight(checkpoints, start_block: Block, end_block: Block) -> float:
    total = 0
    for a, b in zip_longest(list(checkpoints), list(checkpoints)[1:]):
        if a < start_block or a > end_block:
            continue
        b = min(b, end_block) if b else end_block
        total += checkpoints[a] * (b - a) / (end_block - start_block)
    return total


def _get_logs(
    address: Optional[ChecksumAddress],
    topics: Optional[List[str]],
    start: Block,
    end: Block
    ) -> List[LogReceipt]:
    if end - start == BATCH_SIZE - 1:
        response = _get_logs_batch_cached(address, topics, start, end)
    else:
        response = _get_logs_no_cache(address, topics, start, end)
    return response


@auto_retry
def _get_logs_no_cache(
    address: Optional[ChecksumAddress],
    topics: Optional[List[str]],
    start: Block,
    end: Block
    ) -> List[LogReceipt]:
    logger.debug(f'fetching logs {start} to {end}')
    try:
        if address is None:
            response = web3.eth.get_logs({"topics": topics, "fromBlock": start, "toBlock": end})
        elif topics is None:
            response = web3.eth.get_logs({"address": address, "fromBlock": start, "toBlock": end})
        else:
            response = web3.eth.get_logs({"address": address, "topics": topics, "fromBlock": start, "toBlock": end})
    except Exception as e:
        if "Service Unavailable for url:" in str(e) or 'exceed maximum block range' in str(e):
            logger.debug('your node is having trouble, breaking batch in half')
            batch_size = (end - start + 1)
            half_of_batch = batch_size // 2
            batch1_end = start + half_of_batch
            batch2_start = batch1_end + 1
            batch1 = _get_logs_no_cache(address, topics, start, batch1_end)
            batch2 = _get_logs_no_cache(address, topics, batch2_start, end)
            response = batch1 + batch2
        else:
            raise
    return response


@memory.cache()
def _get_logs_batch_cached(
    address: Optional[str],
    topics: Optional[List[str]],
    start: Block,
    end: Block
    ) -> List[LogReceipt]:
    return _get_logs_no_cache(address, topics, start, end)



import logging
from typing import Dict, List, Optional, Tuple

from brownie import chain
from brownie.convert.datatypes import EthAddress
from brownie.exceptions import VirtualMachineError
from y.classes.common import ERC20
from y.classes.singleton import Singleton
from y.constants import dai, usdc, wbtc, weth
from y.contracts import Contract, has_methods
from y.datatypes import UsdPrice, UsdValue
from y.decorators import log
from y.networks import Network
from y.prices import magic
from y.typing import AddressOrContract, AnyAddressType, Block
from y.utils.multicall import fetch_multicall, multicall_decimals
from y.utils.raw_calls import _decimals

EXCHANGE_PROXY = {
    Network.Mainnet: '0x3E66B66Fd1d0b02fDa6C811Da9E0547970DB2f21',
}.get(chain.id, None)

logger = logging.getLogger(__name__)


class BalancerV1Pool(ERC20):
    def __init__(self, pool_address: AnyAddressType) -> None:
        super().__init__(pool_address)

    @log(logger)
    def tokens(self, block: Optional[Block] = None) -> List[ERC20]:
        tokens = self.contract.getCurrentTokens(block_identifier=block)
        return [ERC20(token) for token in tokens]

    @log(logger)
    def get_pool_price(self, block: Optional[Block] = None) -> UsdPrice:
        supply = self.total_supply_readable(block=block)
        if supply == 0:
            return 0
        return UsdPrice(self.get_tvl(block=block) / supply)

    @log(logger)
    def get_tvl(self, block: Optional[Block] = None) -> Optional[UsdValue]:
        token_balances = self.get_balances()
        good_balances = {
            token: balance
            for token, balance
            in token_balances.items()
            if token.price(block=block, return_None_on_failure=True) is not None
        }
        
        # in case we couldn't get prices for all tokens, we can extrapolate from the prices we did get
        good_value = sum(balance * token.price(block=block, return_None_on_failure=True) for token, balance in good_balances.items())
        if len(good_balances):
            return good_value / len(good_balances) * len(token_balances)
        return None

    @log(logger)
    def get_balances(self, block: Optional[Block] = None) -> Dict[ERC20, float]:
        tokens = self.tokens(block=block)
        balances = fetch_multicall(*[[self.contract, "getBalance", token] for token in tokens], block=block)
        balances = [balance if balance else 0 for balance in balances]
        decimals = multicall_decimals(tokens, block)
        return {token:balance / 10 ** decimal for token, balance, decimal in zip(tokens,balances,decimals)}


class BalancerV1(metaclass=Singleton):
    def __init__(self) -> None:
        self.exchange_proxy = Contract(EXCHANGE_PROXY) if EXCHANGE_PROXY else None
    
    def __str__(self) -> str:
        return "BalancerV1()"
    
    @log(logger)
    def is_pool(self, token_address: AnyAddressType) -> bool:
        return has_methods(token_address ,{"getCurrentTokens()(address[])", "getTotalDenormalizedWeight()(uint)", "totalSupply()(uint)"})
    
    @log(logger)
    def get_pool_price(self, token_address: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
        assert self.is_pool(token_address)
        return BalancerV1Pool(token_address).get_pool_price(block=block)

    @log(logger)
    def get_token_price(self, token_address: AddressOrContract, block: Optional[Block] = None) -> UsdPrice:
        out, totalOutput = self.get_some_output(token_address, block=block)
        if out:
            return (totalOutput / 10 ** _decimals(out,block)) * magic.get_price(out, block)
        # Can we get an output if we try smaller size?
        scale = 0.5
        out, totalOutput = self.get_some_output(token_address, block=block, scale=scale) 
        if out:
            return (totalOutput / 10 ** _decimals(out,block)) * magic.get_price(out, block) / scale
        # How about now? 
        scale = 0.1
        out, totalOutput = self.get_some_output(token_address, block=block, scale=scale)
        if out:
            return (totalOutput / 10 ** _decimals(out,block)) * magic.get_price(out, block) / scale
        else:
            return

    @log(logger)
    def check_liquidity_against(
        self,
        token_in: AddressOrContract,
        token_out: AddressOrContract,
        scale: int = 1,
        block: Optional[Block] = None
        ) -> Tuple[EthAddress, int]:

        output = self.exchange_proxy.viewSplitExactIn(
            token_in, token_out, 10 ** _decimals(token_in) * scale, 32 # NOTE: 32 is max
            , block_identifier = block
        )['totalOutput']
        return token_out, output

    @log(logger)
    def get_some_output(
        self,
        token_in: AddressOrContract,
        scale: int = 1,
        block: Optional[Block] = None
        ) -> Tuple[EthAddress,int]:
        try:
            out, totalOutput = self.check_liquidity_against(token_in, weth, block=block)
        except (ValueError, VirtualMachineError):
            try:
                out, totalOutput = self.check_liquidity_against(token_in, dai, block=block)
            except (ValueError, VirtualMachineError):
                try:
                    out, totalOutput = self.check_liquidity_against(token_in, usdc, block=block)
                except (ValueError, VirtualMachineError):
                    try:
                        out, totalOutput = self.check_liquidity_against(token_in, wbtc, block=block)
                    except (ValueError, VirtualMachineError):
                        out = None
                        totalOutput = None
        return out, totalOutput
